#pragma once

#include <ChimeraTK/ApplicationCore/ApplicationCore.h>
#include <tec/TecFrame.h>
#include <tec/TecController.h>
#include <tec/TecChimeraInterface.h>

namespace ctk = ChimeraTK;

struct TecModule : ctk::ApplicationModule {
  TecModule(EntityOwner* owner, const std::string& name, const std::string& description,
      ctk::HierarchyModifier hierarchyModifier = ctk::HierarchyModifier::none,
      const std::unordered_set<std::string>& tags = {});

  ~TecModule() override {}

  struct : ctk::VariableGroup {
    using ctk::VariableGroup::VariableGroup;
    ctk::ScalarPushInput<uint64_t> trigger{this, "tick", "", ""};
  } periodicRead{this, "PeriodicRead", "", ctk::HierarchyModifier::moveToRoot};

  //access to TEC library
  ctk::ScalarOutput<uint32_t> triggerCommands{this, "triggerCommands", "", ""};
  ctk::ScalarPollInput<uint32_t> responseStatus{this, "responseStatus", "", ""};
  ctk::ArrayPollInput<int32_t> commandMemoryRaw_rd{this, "commandMemoryRaw_rd", "", 256, ""};
  ctk::ArrayOutput<int32_t> commandMemoryRaw_wr{this, "commandMemoryRaw_wr", "", 256, ""};

  ctk::ScalarPollInput<uint32_t> tecAddress{
      this, "tecAddress", "", "The address of the TEC device on the serial interface"};

  struct ErrorLog : ctk::VariableGroup {
    using ctk::VariableGroup::VariableGroup;
    ctk::ScalarOutput<std::string> errorMessage{
        this, "errorMessage", "", "Message in case of errors. Need to be reset via errorReset."};
    ctk::ScalarOutput<int32_t> errorStatus{
        this, "errorStatus", "", "Boolean flag whether there is an error. Need to be reset via errorReset."};
    ctk::ScalarPushInput<int32_t> errorReset{this, "errorReset", "", "Clear errorMessage and errorStatus."};

    void append(const std::string& e);
    void reset();
    void write();
  } errorlog{this, "", "", ctk::HierarchyModifier::hideThis};

  ctk::ScalarPushInput<int32_t> errorReset{this, "errorReset", "", "Clear errorMessage and errorStatus."};

  ctk::ScalarPollInput<uint32_t> fpgaTimeout{
      this, "fpgaTimeout", "ms", "Timeout for communication with FPGA firmware."};

  ctk::ScalarPollInput<uint32_t> enableDebug{this, "enableDebug", "", "Enable debug output to console"};

  void mainLoop() override;
  void readPeriodic();
  void readGroup(std::list<TecFrame::TecParameter>& group);
  void writeGroup(std::list<TecFrame::TecParameter>& group);
  void writeGroupToFlash(std::list<TecFrame::TecParameter>& group);
  void writeGroupInternal(std::list<TecFrame::TecParameter>& group, std::vector<TecFrame>& writeRequests);

  /** Perform the actual communication with the TEC device. The list of frames passed to this function is sent to the
   *  device, returned is a list of frames with the replies from the device. If the list exceeds a certain limit,
   *  multiple transfers will be initiated. 
   *
   *  Method returns a list of pointers to the processed responses. All requests that fail due to a timeout is marked as
   *  a nullptr in the return list.
   *  */
  std::vector<std::shared_ptr<TecFrame>> doTecCommunication(std::vector<TecFrame>& requests);

  template<typename T>
  void checkAndUpdateReadResponses(TecFrame* frame, ctk::ScalarOutput<T>& accessor);

  template<typename T>
  void checkAndUpdateWriteResponses(TecFrame* frame, ctk::ScalarOutput<T>& accessor);

  std::list<std::pair<ChimeraTK::ScalarPollInput<int>, TecFrame::TecParameter>> enablePeriodicReadList;
  std::map<TecFrame::TecParameter, ChimeraTK::ScalarPushInputWB<int>> writeIntMap;
  std::map<TecFrame::TecParameter, ChimeraTK::ScalarPushInputWB<float>> writeFloatMap;
  std::map<TecFrame::TecParameter, ChimeraTK::ScalarOutput<int>> readIntMap;
  std::map<TecFrame::TecParameter, ChimeraTK::ScalarOutput<float>> readFloatMap;
  std::map<std::string, ChimeraTK::ScalarPushInput<int>> groupCommandMap;
  std::map<ChimeraTK::TransferElementID, std::list<TecFrame::TecParameter>> writeGroupMap;
  std::map<ChimeraTK::TransferElementID, std::list<TecFrame::TecParameter>> writeGroupToFlashMap;
  std::map<ChimeraTK::TransferElementID, std::list<TecFrame::TecParameter>> readGroupMap;
  std::map<ChimeraTK::TransferElementID, std::list<TecFrame::TecParameter>> setInputFromOutputGroupMap;
  //std::unordered_set<ChimeraTK::TransferElementID> writeBackAccessors;

  void setInputFromOutputGroup(std::list<TecFrame::TecParameter>& group);
  void createReadWriteMaps();
  void createTransferIdGroupMaps();
  void createCommandGroupMaps();
  void createEnablePeriodicList();
  void createWriteBackAccessorsUnorderedSet();

  std::shared_ptr<TecChimeraInterface> tecIface;
  TecController tec{tecIface};

 private:
  std::string getErrorString(TecFrame& f);
};
