// Define a name for the test module.
#define BOOST_TEST_MODULE testErrorCodes
// Only after defining the name include the unit test header.
#include <boost/test/included/unit_test.hpp>
#include <ChimeraTK/ApplicationCore/TestFacility.h>
#include <ChimeraTK/BackendFactory.h>
#include "Server.h"
#include "TecDummy.h"

using namespace boost::unit_test_framework;

struct Fixture_t {
  Server theServer{"tec_dummy"};
  ChimeraTK::TestFacility testFacility;
  boost::shared_ptr<TecDummy> dummy{
      boost::dynamic_pointer_cast<TecDummy>(ChimeraTK::BackendFactory::getInstance().createBackend("ActualDevice"))};
};

/// TestSuite for the server, adapt name
BOOST_AUTO_TEST_SUITE(TecServerTestSuite)

/**********************************************************************************************************************/

/// Test write of an out-of-range value
BOOST_FIXTURE_TEST_CASE(testOutOfRangeWrite, Fixture_t) {
  std::cout << "testOutOfRangeWrite" << std::endl;
  testFacility.runApplication();

  for(int writeToFlash = 0; writeToFlash <= 1; ++writeToFlash) {
    std::cout << "=== testOutOfRangeWrite: writeToFlash = " << writeToFlash << std::endl;

    auto fanControlEnable = testFacility.getScalar<int>("TecModule/ctrl_ch1_fanControlEnable_wr");
    auto fanTargetTemp = testFacility.getScalar<float>("TecModule/ctrl_ch1_fanTargetTemp_wr");
    auto targetObjectTemp = testFacility.getScalar<float>("TecModule/ctrl_ch1_targetObjectTemp_wr");
    auto errorMessage = testFacility.getScalar<std::string>("TecModule/errorMessage");
    auto errorStatus = testFacility.getScalar<int>("TecModule/errorStatus");
    auto errorReset = testFacility.getScalar<int>("TecModule/errorReset");
    ctk::ScalarRegisterAccessor<int> writeCh1Group7;
    ctk::ScalarRegisterAccessor<int> writeCh1Group1;
    if(!writeToFlash) {
      writeCh1Group7.replace(testFacility.getScalar<int>("TecModule/writeCh1Group7_wr"));
      writeCh1Group1.replace(testFacility.getScalar<int>("TecModule/writeCh1Group1_wr"));
    }
    else {
      writeCh1Group7.replace(testFacility.getScalar<int>("TecModule/writeToFlashCh1Group7_wr"));
      writeCh1Group1.replace(testFacility.getScalar<int>("TecModule/writeToFlashCh1Group1_wr"));
    }

    // initially everything should work as normal, no error message must be produced
    fanControlEnable = 1;
    fanControlEnable.write();
    fanTargetTemp = 0.;
    fanTargetTemp.write();
    targetObjectTemp = 0.;
    targetObjectTemp.write();
    writeCh1Group7.write();
    writeCh1Group1.write();
    testFacility.stepApplication();
    BOOST_CHECK_EQUAL(dummy->getInt(TecFrame::TecParameter::ExpFanControlEnable), 1);
    BOOST_CHECK_CLOSE(dummy->getFloat(TecFrame::TecParameter::ExpFanTargetTemp), 0, 0.0001);
    BOOST_CHECK_CLOSE(dummy->getFloat(TecFrame::TecParameter::TemTargetObjectTemp), 0.0, 0.0001);
    BOOST_CHECK(!errorMessage.readNonBlocking());
    BOOST_CHECK(!errorStatus.readNonBlocking());
    BOOST_CHECK(std::string(errorMessage) == "");
    BOOST_CHECK(errorStatus == 0);

    // inject error in dummy
    dummy->errorFlags[TecFrame::TecParameter::ExpFanControlEnable] = TecErrorCode::EER_PAR_OUT_OF_RANGE;
    fanControlEnable = 42;
    fanControlEnable.write();
    writeCh1Group7.write();
    BOOST_CHECK_EQUAL(dummy->getInt(TecFrame::TecParameter::ExpFanControlEnable), 1);
    testFacility.stepApplication();
    BOOST_CHECK(errorMessage.readNonBlocking());
    BOOST_CHECK(errorStatus.readNonBlocking());
    BOOST_CHECK(std::string(errorMessage) != "");
    BOOST_CHECK(std::string(errorMessage).find("ctrl_ch1_fanControlEnable") != std::string::npos);
    BOOST_CHECK(errorStatus != 0);
    BOOST_CHECK_EQUAL(dummy->getInt(TecFrame::TecParameter::ExpFanControlEnable), 1);

    // error needs to be reset manually
    BOOST_CHECK(!errorMessage.readNonBlocking());
    BOOST_CHECK(!errorStatus.readNonBlocking());
    errorReset.write();
    BOOST_CHECK(errorMessage.readNonBlocking());
    BOOST_CHECK(errorStatus.readNonBlocking());
    BOOST_CHECK(std::string(errorMessage) == "");
    BOOST_CHECK(errorStatus == 0);

    // check other variable in different group: no error should be observed
    targetObjectTemp = 3.1415;
    targetObjectTemp.write();
    writeCh1Group1.write();
    BOOST_CHECK_CLOSE(dummy->getFloat(TecFrame::TecParameter::TemTargetObjectTemp), 0.0, 0.0001);
    testFacility.stepApplication();
    BOOST_CHECK_CLOSE(dummy->getFloat(TecFrame::TecParameter::TemTargetObjectTemp), 3.1415, 0.0001);
    BOOST_CHECK(!errorMessage.readNonBlocking());
    BOOST_CHECK(!errorStatus.readNonBlocking());
    BOOST_CHECK(std::string(errorMessage) == "");
    BOOST_CHECK(errorStatus == 0);

    // verify that error is still there for the first variable
    writeCh1Group7.write();
    testFacility.stepApplication();
    BOOST_CHECK(errorMessage.readNonBlocking());
    BOOST_CHECK(errorStatus.readNonBlocking());
    BOOST_CHECK(std::string(errorMessage) != "");
    BOOST_CHECK(std::string(errorMessage).find("ctrl_ch1_fanControlEnable") != std::string::npos);
    BOOST_CHECK(errorStatus != 0);
    BOOST_CHECK_EQUAL(dummy->getInt(TecFrame::TecParameter::ExpFanControlEnable), 1);

    // generate additional error for another variable in the same group: both errors need to be in errorMessage
    dummy->errorFlags[TecFrame::TecParameter::ExpFanTargetTemp] = TecErrorCode::EER_PAR_OUT_OF_RANGE;
    fanTargetTemp = 1500.;
    fanTargetTemp.write();
    writeCh1Group7.write();
    BOOST_CHECK_CLOSE(dummy->getFloat(TecFrame::TecParameter::ExpFanTargetTemp), 0, 0.0001);
    testFacility.stepApplication();
    BOOST_CHECK(errorMessage.readNonBlocking());
    BOOST_CHECK(errorStatus.readNonBlocking());
    BOOST_CHECK(std::string(errorMessage) != "");
    BOOST_CHECK(std::string(errorMessage).find("ctrl_ch1_fanControlEnable") != std::string::npos);
    BOOST_CHECK(std::string(errorMessage).find("ctrl_ch1_fanTargetTemp") != std::string::npos);
    BOOST_CHECK(errorStatus != 0);
    BOOST_CHECK_CLOSE(dummy->getFloat(TecFrame::TecParameter::ExpFanTargetTemp), 0, 0.0001);

    // remove error from dummy and check again. a new value should be immediately visible, but the old still unwritten
    // value should also still come through
    dummy->errorFlags[TecFrame::TecParameter::ExpFanControlEnable] = TecErrorCode::NO_ERROR;
    dummy->errorFlags[TecFrame::TecParameter::ExpFanTargetTemp] = TecErrorCode::NO_ERROR;
    fanControlEnable = -120;
    fanControlEnable.write();
    writeCh1Group7.write();
    testFacility.stepApplication();
    BOOST_CHECK_EQUAL(dummy->getInt(TecFrame::TecParameter::ExpFanControlEnable), -120);
    BOOST_CHECK_CLOSE(dummy->getFloat(TecFrame::TecParameter::ExpFanTargetTemp), 1500., 0.0001);

    // error needs to be reset manually
    BOOST_CHECK(!errorMessage.readNonBlocking());
    BOOST_CHECK(!errorStatus.readNonBlocking());
    errorReset.write();
    BOOST_CHECK(errorMessage.readNonBlocking());
    BOOST_CHECK(errorStatus.readNonBlocking());
    BOOST_CHECK(std::string(errorMessage) == "");
    BOOST_CHECK(errorStatus == 0);
  }
}

/**********************************************************************************************************************/

/// Test read of a value with an error (EER_PAR_NOT_AVAILABLE)
BOOST_FIXTURE_TEST_CASE(testReadOnDemandError, Fixture_t) {
  std::cout << "testReadOnDemandError" << std::endl;
  testFacility.runApplication();
  auto errorMessage = testFacility.getScalar<std::string>("TecModule/errorMessage");
  auto errorStatus = testFacility.getScalar<int>("TecModule/errorStatus");
  auto errorReset = testFacility.getScalar<int>("TecModule/errorReset");
  auto serialNumber = testFacility.getScalar<int>("TecModule/mon_serialNumber_rd");
  auto firmwareVersion = testFacility.getScalar<int>("TecModule/mon_firmwareVersion_rd");

  // Inject error in dummy
  dummy->set(TecFrame::TecParameter::MonSerialNumber, 9);
  dummy->set(TecFrame::TecParameter::MonFirmwareVersion, 98765);
  dummy->errorFlags[TecFrame::TecParameter::MonSerialNumber] = TecErrorCode::EER_PAR_NOT_AVAILABLE;
  testFacility.writeScalar("TecModule/readGroup1_wr", 2);
  testFacility.stepApplication();
  BOOST_CHECK(serialNumber.readNonBlocking());
  BOOST_CHECK(firmwareVersion.readNonBlocking());
  BOOST_CHECK_EQUAL(serialNumber, 0);
  BOOST_CHECK_EQUAL(firmwareVersion, 98765);
  BOOST_CHECK(serialNumber.dataValidity() == ctk::DataValidity::faulty);
  BOOST_CHECK(firmwareVersion.dataValidity() == ctk::DataValidity::ok); // no error for this variable
  BOOST_CHECK(errorMessage.readNonBlocking());
  BOOST_CHECK(errorStatus.readNonBlocking());
  BOOST_CHECK(std::string(errorMessage) != "");
  BOOST_CHECK(std::string(errorMessage).find("mon_serialNumber") != std::string::npos);
  BOOST_CHECK(errorStatus != 0);

  // error needs to be reset manually
  BOOST_CHECK(!errorMessage.readNonBlocking());
  BOOST_CHECK(!errorStatus.readNonBlocking());
  errorReset.write();
  BOOST_CHECK(errorMessage.readNonBlocking());
  BOOST_CHECK(errorStatus.readNonBlocking());
  BOOST_CHECK(std::string(errorMessage) == "");
  BOOST_CHECK(errorStatus == 0);
  BOOST_CHECK(!serialNumber.readNonBlocking()); // DataValidity of variables must not change

  // Clear error in dummy and retry read
  dummy->errorFlags[TecFrame::TecParameter::MonSerialNumber] = TecErrorCode::NO_ERROR;
  testFacility.writeScalar("TecModule/readGroup1_wr", 2);
  testFacility.stepApplication();
  BOOST_CHECK(serialNumber.readNonBlocking());
  BOOST_CHECK(firmwareVersion.readNonBlocking());
  BOOST_CHECK_EQUAL(serialNumber, 9);
  BOOST_CHECK_EQUAL(firmwareVersion, 98765);
  BOOST_CHECK(serialNumber.dataValidity() == ctk::DataValidity::ok);
  BOOST_CHECK(!errorMessage.readNonBlocking());
  BOOST_CHECK(!errorStatus.readNonBlocking());

  // Inject another error in dummy, this time for two variables at the same time
  dummy->set(TecFrame::TecParameter::MonSerialNumber, 12042);
  dummy->set(TecFrame::TecParameter::MonFirmwareVersion, 99);
  dummy->errorFlags[TecFrame::TecParameter::MonSerialNumber] = TecErrorCode::EER_PAR_NOT_AVAILABLE;
  dummy->errorFlags[TecFrame::TecParameter::MonFirmwareVersion] = TecErrorCode::EER_PAR_NOT_AVAILABLE;
  testFacility.writeScalar("TecModule/readGroup1_wr", 2);
  testFacility.stepApplication();
  BOOST_CHECK(serialNumber.readNonBlocking());
  BOOST_CHECK(firmwareVersion.readNonBlocking());
  BOOST_CHECK_EQUAL(serialNumber, 9);
  BOOST_CHECK_EQUAL(firmwareVersion, 98765);
  BOOST_CHECK(serialNumber.dataValidity() == ctk::DataValidity::faulty);
  BOOST_CHECK(firmwareVersion.dataValidity() == ctk::DataValidity::faulty);
  BOOST_CHECK(errorMessage.readNonBlocking());
  BOOST_CHECK(errorStatus.readNonBlocking());
  BOOST_CHECK(std::string(errorMessage) != "");
  BOOST_CHECK(std::string(errorMessage).find("mon_serialNumber") != std::string::npos);
  BOOST_CHECK(std::string(errorMessage).find("mon_firmwareVersion") != std::string::npos);
  BOOST_CHECK(errorStatus != 0);

  // Clear error in dummy for only one variable and retry read
  dummy->errorFlags[TecFrame::TecParameter::MonSerialNumber] = TecErrorCode::NO_ERROR;
  testFacility.writeScalar("TecModule/readGroup1_wr", 2);
  testFacility.stepApplication();
  BOOST_CHECK(serialNumber.readNonBlocking());
  BOOST_CHECK(firmwareVersion.readNonBlocking());
  BOOST_CHECK_EQUAL(serialNumber, 12042);
  BOOST_CHECK_EQUAL(firmwareVersion, 98765);
  BOOST_CHECK(serialNumber.dataValidity() == ctk::DataValidity::ok);
  BOOST_CHECK(firmwareVersion.dataValidity() == ctk::DataValidity::faulty);
  BOOST_CHECK(errorMessage.readNonBlocking());
  BOOST_CHECK(errorStatus.readNonBlocking());
  BOOST_CHECK(std::string(errorMessage) != "");
  BOOST_CHECK(std::string(errorMessage).find("mon_serialNumber") == std::string::npos);
  BOOST_CHECK(std::string(errorMessage).find("mon_firmwareVersion") != std::string::npos);
  BOOST_CHECK(errorStatus != 0);

  // Clear error in dummy for second variable and retry read
  dummy->errorFlags[TecFrame::TecParameter::MonFirmwareVersion] = TecErrorCode::NO_ERROR;
  testFacility.writeScalar("TecModule/readGroup1_wr", 2);
  testFacility.stepApplication();
  BOOST_CHECK(serialNumber.readNonBlocking());
  BOOST_CHECK(firmwareVersion.readNonBlocking());
  BOOST_CHECK_EQUAL(serialNumber, 12042);
  BOOST_CHECK_EQUAL(firmwareVersion, 99);
  BOOST_CHECK(serialNumber.dataValidity() == ctk::DataValidity::ok);
  BOOST_CHECK(firmwareVersion.dataValidity() == ctk::DataValidity::ok);
  BOOST_CHECK(!errorMessage.readNonBlocking());
  BOOST_CHECK(!errorStatus.readNonBlocking());

  // error needs to be reset manually
  errorReset.write();
  BOOST_CHECK(errorMessage.readNonBlocking());
  BOOST_CHECK(errorStatus.readNonBlocking());
  BOOST_CHECK(std::string(errorMessage) == "");
  BOOST_CHECK(errorStatus == 0);
}

/**********************************************************************************************************************/

/// Test periodic read of a value with an error (EER_PAR_NOT_AVAILABLE)
BOOST_FIXTURE_TEST_CASE(testReadPeriodicError, Fixture_t) {
  std::cout << "testReadPeriodicError" << std::endl;
  testFacility.runApplication();
  auto errorMessage = testFacility.getScalar<std::string>("TecModule/errorMessage");
  auto errorStatus = testFacility.getScalar<int>("TecModule/errorStatus");
  auto errorReset = testFacility.getScalar<int>("TecModule/errorReset");
  auto serialNumber = testFacility.getScalar<int>("TecModule/mon_serialNumber_rd");

  testFacility.writeScalar<int>("TecModule/mon_serialNumber_enaPeriodicRead", 1);

  dummy->set(TecFrame::TecParameter::MonSerialNumber, 17);
  dummy->errorFlags[TecFrame::TecParameter::MonSerialNumber] = TecErrorCode::EER_PAR_NOT_AVAILABLE;
  theServer.periodicRead.sendTrigger();
  testFacility.stepApplication();
  BOOST_CHECK(serialNumber.readNonBlocking());
  BOOST_CHECK_EQUAL(serialNumber, 0);
  BOOST_CHECK(serialNumber.dataValidity() == ctk::DataValidity::faulty);
  BOOST_CHECK(errorMessage.readNonBlocking());
  BOOST_CHECK(errorStatus.readNonBlocking());
  BOOST_CHECK(std::string(errorMessage) != "");
  BOOST_CHECK(std::string(errorMessage).find("mon_serialNumber") == std::string::npos);
  BOOST_CHECK(errorStatus != 0);

  theServer.periodicRead.sendTrigger();
  testFacility.stepApplication();
  BOOST_CHECK(serialNumber.readNonBlocking());
  BOOST_CHECK_EQUAL(serialNumber, 0);
  BOOST_CHECK(serialNumber.dataValidity() == ctk::DataValidity::faulty);
  BOOST_CHECK(errorMessage.readNonBlocking());
  BOOST_CHECK(errorStatus.readNonBlocking());
  BOOST_CHECK(std::string(errorMessage) != "");
  BOOST_CHECK(std::string(errorMessage).find("mon_serialNumber") == std::string::npos);
  BOOST_CHECK(errorStatus != 0);

  // error needs to be reset manually
  BOOST_CHECK(!errorMessage.readNonBlocking());
  BOOST_CHECK(!errorStatus.readNonBlocking());
  errorReset.write();
  BOOST_CHECK(errorMessage.readNonBlocking());
  BOOST_CHECK(errorStatus.readNonBlocking());
  BOOST_CHECK(std::string(errorMessage) == "");
  BOOST_CHECK(errorStatus == 0);

  theServer.periodicRead.sendTrigger();
  testFacility.stepApplication();
  BOOST_CHECK(serialNumber.readNonBlocking());
  BOOST_CHECK_EQUAL(serialNumber, 0);
  BOOST_CHECK(serialNumber.dataValidity() == ctk::DataValidity::faulty);
  BOOST_CHECK(errorMessage.readNonBlocking());
  BOOST_CHECK(errorStatus.readNonBlocking());
  BOOST_CHECK(std::string(errorMessage) != "");
  BOOST_CHECK(std::string(errorMessage).find("mon_serialNumber") == std::string::npos);
  BOOST_CHECK(errorStatus != 0);

  dummy->errorFlags[TecFrame::TecParameter::MonSerialNumber] = TecErrorCode::NO_ERROR;
  theServer.periodicRead.sendTrigger();
  testFacility.stepApplication();
  BOOST_CHECK(serialNumber.readNonBlocking());
  BOOST_CHECK_EQUAL(serialNumber, 17);
  BOOST_CHECK(serialNumber.dataValidity() == ctk::DataValidity::ok);

  // error needs to be reset manually
  BOOST_CHECK(!errorMessage.readNonBlocking());
  BOOST_CHECK(!errorStatus.readNonBlocking());
  errorReset.write();
  BOOST_CHECK(errorMessage.readNonBlocking());
  BOOST_CHECK(errorStatus.readNonBlocking());
  BOOST_CHECK(std::string(errorMessage) == "");
  BOOST_CHECK(errorStatus == 0);
}

/**********************************************************************************************************************/

BOOST_AUTO_TEST_SUITE_END()
